#!/usr/bin/bash

# Incus Container Import Script
#
# This script imports an Incus instance from a remote server or updates an existing local instance with remote config and data.
#
# Version: V1.1.0 (2025-03-20)

# Function to log messages
log_msg() {
    LOG_FILE="/var/log/incus_import.log"
    echo "$(date --rfc-3339=seconds) - $1" | tee -a "$LOG_FILE"
}

# Check SSH connectivity to source server
check_ssh_connection() {
    local server=$1
    local ssh_port=${SSH_PORT:-22}
    local timeout=${SSH_TIMEOUT:-10}

    if ! ssh -p "$ssh_port" -o BatchMode=yes -o ConnectTimeout="$timeout" root@"$server" "true" 2>/dev/null; then
        log_msg "ERROR: Unable to connect to $server via SSH on port $ssh_port. Aborting."
        exit 1
    fi
    return 0
}

# Ensure BlueOnyx image server is added (only for BlueOnyx images)
ensure_blueonyx_image_server() {
    local remote_name="blueonyx-images"
    local remote_url="https://incus.aventurin.net:443/"
    
    if ! incus remote list | grep -q "^|[[:space:]]*$remote_name[[:space:]]"; then
        log_msg "Adding BlueOnyx image server ($remote_url) ..."
        if ! incus remote add "$remote_name" "$remote_url" --protocol=simplestreams --public --accept-certificate --debug; then
            log_msg "Error: Failed to add BlueOnyx image server ($remote_url). Check network or configuration."
            return 1
        fi
        log_msg "BlueOnyx image server added successfully."
    else
        log_msg "BlueOnyx image server ($remote_name) already configured."
    fi
    return 0
}

# Fetch image (BlueOnyx from custom server, others from default images:)
fetch_image() {
    local image_alias=$1
    local remote_name="blueonyx-images"
    local check_alias="$image_alias"

    # Adjust check_alias for Debian (remove /amd64 for local check)
    if [[ "$image_alias" =~ ^debian/ ]]; then
        check_alias=$(echo "$image_alias" | sed 's|/amd64$||')
    fi

    if [[ "$image_alias" =~ ^blueonyx/ ]]; then
        ensure_blueonyx_image_server || return 1
        case "$image_alias" in
            blueonyx-5210R-cloud) image_alias="blueonyx/5210R/cloud" ;;
            blueonyx-5211R-cloud) image_alias="blueonyx/5211R/cloud" ;;
        esac
        check_alias="$image_alias"
        if ! incus image list | grep -q "$check_alias"; then
            log_msg "Image $image_alias not found locally. Attempting to copy from $remote_name ..."
            if ! incus image copy "$remote_name:$image_alias" local: --alias "$image_alias"; then
                log_msg "Error: Failed to fetch image $image_alias from $remote_name."
                return 1
            fi
            log_msg "Image $image_alias copied successfully from $remote_name!"
        else
            log_msg "Image $check_alias already available locally."
        fi
    else
        if ! incus image list | grep -q "$check_alias"; then
            log_msg "Image $image_alias not found locally. Attempting to copy from images: ..."
            if ! incus image copy "images:$image_alias" local: --alias "$check_alias"; then
                log_msg "Error: Failed to fetch image $image_alias from images:."
                return 1
            fi
            log_msg "Image $check_alias copied successfully from images:!"
        else
            log_msg "Image $check_alias already available locally."
        fi
    fi
    return 0
}

# Create or update the local container
create_container() {
    local container_name=$1
    local source_server=$2

    if ! check_ssh_connection "$source_server"; then
        log_msg "ERROR: Cannot establish SSH connection to $source_server. Aborting container creation for $container_name."
        return 1
    fi

    # Fetch the config from the source server
    log_msg "Fetching configuration for $container_name from $source_server ..."
    ssh -p "$SSH_PORT" root@"$source_server" "incus config show $container_name" > "/tmp/$container_name.yaml" || {
        log_msg "Error: Failed to fetch configuration for $container_name from $source_server."
        return 1
    }

    # Extract image information
    IMAGE_OS=$(grep 'image.os:' "/tmp/$container_name.yaml" | awk '{print tolower($2)}')
    IMAGE_RELEASE=$(grep 'image.release:' "/tmp/$container_name.yaml" | awk '{print $2}' | sed 's/"//g')
    IMAGE_VARIANT=$(grep 'image.release:' "/tmp/$container_name.yaml" | awk '{print $2}' | sed 's/"//g')
    IMAGE_ARCH="amd64" # Force architecture to amd64 for simplicity

    # Determine image alias
    if [[ "$IMAGE_OS" == "blueonyx" ]]; then
        PARTIAL_ALIAS="blueonyx/${IMAGE_RELEASE}/cloud"
    else
        if [ -z "$IMAGE_OS" ] && [ -z "$IMAGE_RELEASE" ]; then
            log_msg "No image metadata found. Assuming blueonyx/5210R/cloud."
            PARTIAL_ALIAS="blueonyx/5210R/cloud"
        else
            # Map Debian releases to versions
            declare -A DEBIAN_RELEASE_MAP=( ["buster"]="10" ["bullseye"]="11" ["bookworm"]="12" ["trixie"]="13" )
            if [[ "$IMAGE_OS" == "debian" ]]; then
                IMAGE_RELEASE=${DEBIAN_RELEASE_MAP[$IMAGE_RELEASE]}
            fi
            PARTIAL_ALIAS="${IMAGE_OS}/${IMAGE_RELEASE}/cloud/${IMAGE_ARCH}"
        fi
    fi

    # Ensure the image exists locally
    if ! fetch_image "$PARTIAL_ALIAS"; then
        log_msg "Error: Image fetch failed for $PARTIAL_ALIAS. Cannot proceed."
        return 1
    fi

    if ! incus list | grep -q "^| $container_name\s"; then
        log_msg "Creating $container_name with image $PARTIAL_ALIAS"
        incus init "$PARTIAL_ALIAS" "$container_name" || {
            log_msg "Error: Failed to initialize $container_name with image $PARTIAL_ALIAS."
            return 1
        }
    else
        log_msg "Container $container_name already exists. Proceeding with config update."
    fi

    # Convert YAML config to JSON for easier parsing
    python3 -c "import sys, yaml, json; yaml_data = yaml.safe_load(sys.stdin); json.dump(yaml_data, sys.stdout, indent=2)" < "/tmp/$container_name.yaml" > "/tmp/$container_name.json" || {
        log_msg "Error: Failed to convert YAML config to JSON for $container_name."
        return 1
    }

    # Apply configuration
    log_msg "Applying configuration settings for $container_name ..."

    # Simple key-value pairs to apply (excluding boot.autostart)
    simple_config_keys=(
        "boot.autostart.delay"
        "boot.autostart.priority"
        "limits.cpu"
        "limits.cpu.priority"
        "limits.memory"
        "security.privileged"
        "user.instance_owner"
        "user.network-config"
        "user.user-data"
        "user.instance_ipv4"
    )

    for key in "${simple_config_keys[@]}"; do
        value=$(python3 -c "import json; data = json.load(open('/tmp/$container_name.json')); print(data['config'].get('$key', ''))")
        if [ -n "$value" ]; then
            log_msg "Setting $key to $value"
            incus config set "$container_name" "$key" "$value" || {
                log_msg "Error: Failed to set $key for $container_name"
                return 1
            }
        fi
    done

    # Force specific settings
    log_msg "Setting security.protection.delete to false"
    incus config set "$container_name" "security.protection.delete" "false"

    # Apply devices
    log_msg "Applying devices for $container_name ..."

    # Handle eth0 device
    current_eth0_config=$(incus config device show "$container_name" eth0 2>/dev/null | sed 's/^ *//;s/: /=/g' | tr '\n' ' ' | sed 's/  / /g')
    eth0_config=$(python3 -c "import json; data = json.load(open('/tmp/$container_name.json')); eth0_config = data['devices'].get('eth0', {}); print(' '.join([f'{k}={v}' for k, v in eth0_config.items()]))")
    if [ "$current_eth0_config" != "$eth0_config" ] && [ -n "$eth0_config" ]; then
        log_msg "Updating eth0 device with config: $eth0_config"
        incus config device remove "$container_name" eth0 2>/dev/null
        incus config device override "$container_name" eth0 $eth0_config || {
            log_msg "Error: Failed to override eth0 device for $container_name"
            return 1
        }
    fi

    # Handle root device
    current_root_config=$(incus config device show "$container_name" root 2>/dev/null | sed 's/^ *//;s/: /=/g' | tr '\n' ' ' | sed 's/  / /g')
    root_config=$(python3 -c "import json; data = json.load(open('/tmp/$container_name.json')); root_config = data['devices'].get('root', {}); print(' '.join([f'{k}={v}' for k, v in root_config.items()]))")
    if [ "$current_root_config" != "$root_config" ] && [ -n "$root_config" ]; then
        log_msg "Updating root device with config: $root_config"
        incus config device remove "$container_name" root 2>/dev/null
        incus config device override "$container_name" root $root_config || {
            log_msg "Error: Failed to override root device for $container_name"
            return 1
        }
    fi

    rm -f "/tmp/$container_name.yaml" "/tmp/$container_name.json"
}

# Sync container data via rsync with two-pass approach for running instances, unless --nofreeze is specified
sync_container_data() {
    local container_name=$1
    local source_server=$2

    log_msg "Syncing data for container $container_name ..."

    check_ssh_connection "$source_server"

    RSYNC_OPTIONS="-ah --info=progress2 --partial --delete --links --hard-links --perms --owner --group --devices --times"
    SSH_CMD="ssh -p $SSH_PORT -c aes128-ctr -o Compression=no"
    SOURCE_PATH="root@$source_server:/home/incus/storage-pools/default/containers/$container_name/rootfs/"
    DEST_PATH="/home/incus/storage-pools/default/containers/$container_name/rootfs/"

    # Check if source is running
    IS_RUNNING_ON_SOURCE=$(ssh -p "$SSH_PORT" root@"$source_server" "incus list $container_name | grep -c RUNNING")

    if [ "$NOFREEZE" -eq 1 ]; then
        # Single rsync with no freeze, ignoring inconsistencies
        log_msg "Using --nofreeze: Performing single rsync (inconsistencies may occur) ..."
        rsync $RSYNC_OPTIONS -e "$SSH_CMD" "$SOURCE_PATH" "$DEST_PATH"
        log_msg "Single rsync completed (no consistency guarantee due to --nofreeze)."
    elif [ "$IS_RUNNING_ON_SOURCE" -eq 1 ]; then
        # First pass: rsync without error checking for running instance
        log_msg "Source $container_name is running. Performing initial rsync (may have inconsistencies) ..."
        rsync $RSYNC_OPTIONS -e "$SSH_CMD" "$SOURCE_PATH" "$DEST_PATH"
        log_msg "Initial rsync completed. Freezing source $container_name for consistent sync ..."

        # Freeze the source instance
        ssh -p "$SSH_PORT" root@"$source_server" "incus freeze $container_name" || {
            log_msg "Error: Failed to freeze $container_name on source server $source_server"
            return 1
        }

        # Second pass: rsync with error checking while frozen
        log_msg "Performing final rsync with $container_name frozen ..."
        rsync $RSYNC_OPTIONS -e "$SSH_CMD" "$SOURCE_PATH" "$DEST_PATH" || {
            ssh -p "$SSH_PORT" root@"$source_server" "incus unfreeze $container_name" # Ensure unfreeze even on failure
            log_msg "Error: Final rsync failed for $container_name"
            return 1
        }

        # Unfreeze the source instance
        log_msg "Unfreezing source $container_name ..."
        ssh -p "$SSH_PORT" root@"$source_server" "incus unfreeze $container_name" || {
            log_msg "Error: Failed to unfreeze $container_name on source server $source_server"
            return 1
        }
    else
        # Single pass: rsync with error checking for stopped instance
        log_msg "Source $container_name is stopped. Performing single rsync with consistency check ..."
        rsync $RSYNC_OPTIONS -e "$SSH_CMD" "$SOURCE_PATH" "$DEST_PATH" || {
            log_msg "Error: Rsync failed for $container_name"
            return 1
        }
    fi

    log_msg "Data sync completed successfully for $container_name."
}

# Main migration function
migrate_container() {
    local container_name=$1
    local source_server=$2

    log_msg "Starting import of Incus instance '$container_name' from $source_server ..."

    check_ssh_connection "$source_server"

    # Check if source container exists and is running
    IS_RUNNING_ON_SOURCE=$(ssh -p "$SSH_PORT" root@"$source_server" "incus list $container_name | grep -c RUNNING")
    if [ "$IS_RUNNING_ON_SOURCE" -eq 0 ]; then
        log_msg "Container $container_name is not running on the source server $source_server. Skipping migration."
        return 0
    fi

    # Capture the original boot.autostart value from the source before any changes
    SOURCE_BOOT_AUTOSTART=$(ssh -p "$SSH_PORT" root@"$source_server" "incus config get $container_name boot.autostart")
    if [ "$SOURCE_BOOT_AUTOSTART" = "true" ]; then
        SOURCE_HAD_AUTOSTART=1
    else
        SOURCE_HAD_AUTOSTART=0
    fi
    log_msg "Original source $container_name boot.autostart is '$SOURCE_BOOT_AUTOSTART'"

    # Check if local container is running
    IS_RUNNING_LOCAL=$(incus list "$container_name" | grep -c RUNNING)
    if [ "$IS_RUNNING_LOCAL" -eq 1 ]; then
        log_msg "Container $container_name is running locally. Stopping it for migration ..."
        incus stop "$container_name" || {
            log_msg "Error: Failed to stop local container $container_name"
            return 1
        }
    fi

    create_container "$container_name" "$source_server" || return 1
    sync_container_data "$container_name" "$source_server" || return 1

    # Handle --start option
    if [ "$START_INSTANCE" -eq 1 ]; then
        log_msg "Setting boot.autostart to false on source server $source_server for $container_name ..."
        ssh -p "$SSH_PORT" root@"$source_server" "incus config set $container_name boot.autostart false" || {
            log_msg "Error: Failed to set boot.autostart to false on source server $source_server for $container_name"
            return 1
        }

        log_msg "Stopping $container_name on source server $source_server ..."
        ssh -p "$SSH_PORT" root@"$source_server" "incus stop $container_name" || {
            log_msg "Error: Failed to stop $container_name on source server $source_server"
            return 1
        }

        if [ "$SOURCE_HAD_AUTOSTART" -eq 1 ]; then
            log_msg "Setting boot.autostart to true on local instance $container_name (mirroring original source setting) ..."
            incus config set "$container_name" "boot.autostart" "true" || {
                log_msg "Error: Failed to set boot.autostart to true on local instance $container_name"
                return 1
            }
        else
            log_msg "Leaving boot.autostart unset on local instance $container_name (source was not set to true)"
        fi

        log_msg "Starting $container_name locally ..."
        incus start "$container_name" || {
            log_msg "Error: Failed to start $container_name locally"
            return 1
        }
    else
        if [ "$SOURCE_HAD_AUTOSTART" -eq 1 ]; then
            log_msg "Note: Source had boot.autostart true, but --start not specified; local instance $container_name left stopped."
        else
            log_msg "Local instance $container_name left stopped (use --start to activate)."
        fi
    fi

    log_msg "Migration completed successfully for $container_name."
}

# Parse CLI arguments
START_INSTANCE=0  # Default to not starting the instance
NOFREEZE=0        # Default to using freeze for running instances
while [[ $# -gt 0 ]]; do
    case "$1" in
        --source-server)
            SOURCE_SERVER="$2"
            shift 2
            ;;
        --container-name)
            CONTAINER_NAME="$2"
            shift 2
            ;;
        --ssh-port)
            SSH_PORT="$2"
            shift 2
            ;;
        --start)
            START_INSTANCE=1
            shift
            ;;
        --nofreeze)
            NOFREEZE=1
            shift
            ;;
        *)
            log_msg "Unknown option: $1"
            echo ""
            echo "Usage: $0 --source-server <SERVER> --container-name <NAME> [--ssh-port <PORT>] [--start] [--nofreeze]"
            echo ""
            echo "    --source-server   IP or FQDN of the source server where the Instance resides that you want to import"
            echo "    --container-name  Name of the instance you want to import."
            echo "    --ssh-port        (Optional) SSH port. If not given it assumes port 22."
            echo "    --start           (Optional) Stop the Instance on the source server (and set 'autostart' to 'false') and start it here."
            echo "    --nofreeze        (Optional) For consistency a 2nd RSYNC pass is run while the source container is frozen. This switch "
            echo "                                 disables the freeze, but this may result in data inconsistencies in the imported instance."
            exit 1
            ;;
    esac
done

# Default SSH port if not specified
SSH_PORT=${SSH_PORT:-22}

# Validate required parameters
if [[ -z "$SOURCE_SERVER" || -z "$CONTAINER_NAME" ]]; then
    log_msg "Error: Missing required parameters!"
    echo ""
    echo "Usage: $0 --source-server <SERVER> --container-name <NAME> [--ssh-port <PORT>] [--start] [--nofreeze]"
    echo ""
    echo "    --source-server   IP or FQDN of the source server where the Instance resides that you want to import"
    echo "    --container-name  Name of the instance you want to import."
    echo "    --ssh-port        (Optional) SSH port. If not given it assumes port 22."
    echo "    --start           (Optional) Stop the Instance on the source server (and set 'autostart' to 'false') and start it here."
    echo "    --nofreeze        (Optional) For consistency a 2nd RSYNC pass is run while the source container is frozen. This switch "
    echo "                                 disables the freeze, but this may result in data inconsistencies in the imported instance."
    exit 1
fi

# Execute the migration
migrate_container "$CONTAINER_NAME" "$SOURCE_SERVER"

exit 0

# 
# Copyright (c) 2006-2025 Michael Stauber, SOLARSPEED.NET
# Copyright (c) 2006-2025 Team BlueOnyx, BLUEONYX.IT
# All Rights Reserved.
# 
# 1. Redistributions of source code must retain the above copyright 
#     notice, this list of conditions and the following disclaimer.
# 
# 2. Redistributions in binary form must reproduce the above copyright 
#     notice, this list of conditions and the following disclaimer in 
#     the documentation and/or other materials provided with the 
#     distribution.
# 
# 3. Neither the name of the copyright holder nor the names of its 
#     contributors may be used to endorse or promote products derived 
#     from this software without specific prior written permission.
# 
# THIS SOFTWARE IS PROVIDED BY THE COPYRIGHT HOLDERS AND CONTRIBUTORS 
# "AS IS" AND ANY EXPRESS OR IMPLIED WARRANTIES, INCLUDING, BUT NOT 
# LIMITED TO, THE IMPLIED WARRANTIES OF MERCHANTABILITY AND FITNESS 
# FOR A PARTICULAR PURPOSE ARE DISCLAIMED. IN NO EVENT SHALL THE 
# COPYRIGHT HOLDER OR CONTRIBUTORS BE LIABLE FOR ANY DIRECT, INDIRECT, 
# INCIDENTAL, SPECIAL, EXEMPLARY, OR CONSEQUENTIAL DAMAGES (INCLUDING, 
# BUT NOT LIMITED TO, PROCUREMENT OF SUBSTITUTE GOODS OR SERVICES; 
# LOSS OF USE, DATA, OR PROFITS; OR BUSINESS INTERRUPTION) HOWEVER 
# CAUSED AND ON ANY THEORY OF LIABILITY, WHETHER IN CONTRACT, STRICT 
# LIABILITY, OR TORT (INCLUDING NEGLIGENCE OR OTHERWISE) ARISING IN 
# ANY WAY OUT OF THE USE OF THIS SOFTWARE, EVEN IF ADVISED OF THE 
# POSSIBILITY OF SUCH DAMAGE.
# 
# You acknowledge that this software is not designed or intended for 
# use in the design, construction, operation or maintenance of any 
# nuclear facility.
#