#!/usr/bin/bash

# OpenVZ 7 CT to Incus Migration Script
#
# This script can import OpenVZ 7 CTs (Debian 10/11/12/13 or BlueOnyx 5210R) straight to Incus or Aventurin{e} 6110R
#
# Version: V1.5.2 (2025-03-10)

# Check SSH connectivity to source server
check_ssh_connection() {
    local server=$1
    local ssh_port=${SSH_PORT:-22}
    local timeout=${SSH_TIMEOUT:-10}

    if ! ssh -p "$ssh_port" -o BatchMode=yes -o ConnectTimeout="$timeout" root@"$server" "true" 2>/dev/null; then
        log_msg "ERROR: Unable to connect to $server via SSH on port $ssh_port. Aborting."
        exit 1
    fi
    return 0
}

# Function to log messages
log_msg() {
    LOG_FILE="/var/log/incus_import.log"
    echo "$(date --rfc-3339=seconds) - $1" | tee -a "$LOG_FILE"
}

# Default values
SSH_PORT=22
SSH_TIMEOUT=10

# Parse named parameters
while [[ $# -gt 0 ]]; do
    case "$1" in
        --vpsid)
            SOURCE_VPSID="$2"
            shift 2
            ;;
        --node)
            SOURCE_SERVER="$2"
            shift 2
            ;;
        --port)
            SSH_PORT="$2"
            shift 2
            ;;
        --name)
            CONTAINER_NAME="$2"
            shift 2
            ;;
        --gateway)
            GATEWAY="$2"
            shift 2
            ;;
        *)
            echo "Unknown option: $1"
            echo "Usage: $0 --vpsid <VPS_ID> --node <NODE> --port <SSH_PORT> --name <CONTAINER_NAME> --gateway <GATEWAY>"
            exit 1
            ;;
    esac
done

# Validate required parameters
if [[ -z "$SOURCE_VPSID" || -z "$SOURCE_SERVER" || -z "$CONTAINER_NAME" || -z "$GATEWAY" ]]; then
    echo "This script can migrate OpenVZ 7 CTs (Debian 10/11/12/13 or BlueOnyx 5210R) to Incus."
    echo "Error: Missing required parameters!"
    echo "Usage: $0 --vpsid <VPS_ID> --node <NODE> --port <SSH_PORT> --name <CONTAINER_NAME> --gateway <GATEWAY>"
    exit 1
fi

# *** Ported from common.sh: BlueOnyx Image Management ***
ensure_blueonyx_image_server() {
    local remote_name="blueonyx-images"
    local remote_url="https://incus.aventurin.net:443/"
    
    if ! incus remote list | grep -q "^|[[:space:]]*$remote_name[[:space:]]"; then
        log_msg "Adding BlueOnyx image server..."
        incus remote add "$remote_name" "$remote_url" --protocol=simplestreams --public --accept-certificate --debug
    fi
}

fetch_blueonyx_image() {
    local image_alias=$1
    local remote_name="blueonyx-images"

    ensure_blueonyx_image_server

    case "$image_alias" in
        blueonyx-5210R-cloud) image_alias="blueonyx/5210R/cloud" ;;
        blueonyx-5211R-cloud) image_alias="blueonyx/5211R/cloud" ;;
    esac

    if ! incus image list | grep -q "$image_alias"; then
        log_msg "Image $image_alias not found locally. Attempting to copy from $remote_name..."
        if ! incus image copy "$remote_name:$image_alias" local: --alias "$image_alias"; then
            log_msg "Error: Failed to fetch image $image_alias from $remote_name."
            return 1
        fi
    fi
    log_msg "Image $image_alias copied successfully!"
}

# *** Enhanced from common.sh: Extract OpenVZ CT details with DNS_SERVERS at the end ***
extract_openvz_config() {
    SOURCE_VPSID=$1

    log_msg "Extracting OpenVZ configuration for CT $SOURCE_VPSID..." >&2

    check_ssh_connection "$SOURCE_SERVER"

    CT_INFO=$(ssh -p "$SSH_PORT" root@"$SOURCE_SERVER" "prlctl list $SOURCE_VPSID --info") || { 
        log_msg "Error: Unable to fetch container info for $SOURCE_VPSID" >&2 
        return 1
    }

    CT_STATE=$(echo "$CT_INFO" | grep "State:" | awk '{print $2}')
    if [[ "$CT_STATE" != "running" && "$CT_STATE" != "mounted" ]]; then
        log_msg "Error: CT $SOURCE_VPSID is not in a valid state (running or mounted). Current state: $CT_STATE" >&2
        return 1
    fi

    HOME_PATH=$(echo "$CT_INFO" | grep "Home:" | cut -d' ' -f2- | tr -d ' ')
    ENV_ID=$(echo "$CT_INFO" | grep "EnvID:" | cut -d' ' -f2- | tr -d ' ')

    if [[ -z "$ENV_ID" ]]; then
        log_msg "Fatal: ENV_ID could not be determined for CT $SOURCE_VPSID. Aborting to prevent unintended rsync behavior!" >&2
        exit 1
    fi

    if [[ -z "$HOME_PATH" || -z "$ENV_ID" ]]; then
        log_msg "Error: Unable to determine Home path or EnvID for CT $SOURCE_VPSID" >&2
        return 1
    fi

    VE_CONF="$HOME_PATH/ve.conf"
    check_ssh_connection "$SOURCE_SERVER"
    CONFIG=$(ssh -p "$SSH_PORT" root@"$SOURCE_SERVER" "cat $VE_CONF") || {
        log_msg "Error: Unable to fetch ve.conf for CT $SOURCE_VPSID" >&2
        return 1
    }

    DISK_MB=$(echo "$CONFIG" | grep -oP 'DISKSPACE=("?)\K\d+(?=:\d+\1)' | tr -d '[:space:]')
    RAM_MB=$(echo "$CONFIG" | grep -oP 'PHYSPAGES="\K[0-9]+(:[0-9]+)?' | sed 's/^0://' | cut -d':' -f1 || echo "8192")
    RAM_MB=$((RAM_MB * 4))

    IP_RAW=$(echo "$CONFIG" | grep -oP 'IP_ADDRESS="\K[^"]+' | awk '{print $1}')
    IP_ADDRESS=$(echo "$IP_RAW" | cut -d'/' -f1)
    NETMASK=$(echo "$IP_RAW" | cut -d'/' -f2)
    if [[ -z "$NETMASK" ]]; then
        NETMASK="255.255.255.0"
    fi

    HOSTNAME=$(echo "$CONFIG" | grep -oP 'HOSTNAME="\K[^"]+')
    OSTEMPLATE=$(echo "$CONFIG" | grep -oP 'OSTEMPLATE="\K[^"]+' | tr -d '[:space:]')

    if [[ "$OSTEMPLATE" =~ ^debian- ]] || [[ "$OSTEMPLATE" =~ ^.debian- ]]; then
        OS_TYPE="debian"
        log_msg "Info: CT is a Debian" >&2
        
        check_ssh_connection "$SOURCE_SERVER"
        if [[ "$CT_STATE" == "running" ]]; then
            OS_VERSION=$(ssh -p "$SSH_PORT" root@"$SOURCE_SERVER" "prlctl exec $SOURCE_VPSID cat /etc/debian_version" | cut -d'.' -f1)
        else
            OS_VERSION=$(ssh -p "$SSH_PORT" root@"$SOURCE_SERVER" "cat /vz/root/$ENV_ID/etc/debian_version" | cut -d'.' -f1)
        fi

        log_msg "Info: CT is a Debian $OS_VERSION" >&2

        if [[ -z "$OS_VERSION" ]]; then
            log_msg "Error: Could not determine Debian version from /etc/debian_version." >&2
            return 1
        fi
    elif [[ "$OSTEMPLATE" =~ ^blueonyx-([0-9]+) ]] || [[ "$OSTEMPLATE" =~ ^.blueonyx-([0-9]+) ]]; then
        OS_TYPE="blueonyx"
        OS_VERSION="5210R"
        log_msg "Info: CT is a BlueOnyx" >&2
    else
        log_msg "Error: Unsupported OS type '$OSTEMPLATE' in $VE_CONF" >&2
        return 1
    fi

    # Extract DNS servers and move to the end (comma-separated)
    DNS_SERVERS=$(echo "$CONFIG" | grep -oP 'NAMESERVER="\K[^"]+' | tr '\n' ',' | sed 's/,$//')

    # Output with DNS_SERVERS at the end
    echo "$RAM_MB $DISK_MB $IP_ADDRESS $NETMASK $HOSTNAME $OS_TYPE $OS_VERSION $ENV_ID $VE_CONF $DNS_SERVERS"
}

# Function to generate VLAN ID (ensures uniqueness)
generate_vlan_id() {
    log_msg "Generating unique VLAN ID..."
    local vlan_id=2
    while incus list --format=json | jq -r '.[].devices.eth0.vlan' | grep -qw "$vlan_id"; do
        ((vlan_id++))
    done
    log_msg "Assigned VLAN ID: $vlan_id"
    echo "$vlan_id"
}

# *** Enhanced from common.sh: Create Incus instance ***
create_container() {
    SOURCE_VPSID=$1
    CONTAINER_NAME=$2

    if [ -z "$GATEWAY" ]; then
        log_msg "Error: GATEWAY variable is not set in config. Aborting."
        return 1
    fi

    log_msg "Creating container $CONTAINER_NAME for OpenVZ CT $SOURCE_VPSID..."

    CONFIG=$(extract_openvz_config "$SOURCE_VPSID" 2>>"$LOG_FILE")
    if [ $? -ne 0 ]; then
        log_msg "Error: Failed to extract configuration for CT $SOURCE_VPSID"
        return 1
    fi

    log_msg "DEBUG: Extracted CONFIG='$CONFIG'"

    # Adjusted to match new order with DNS_SERVERS at the end
    IFS=' ' read -r RAM_MB DISK_MB IP_ADDRESS NETMASK HOSTNAME OS_TYPE OS_VERSION ENV_ID VE_CONF DNS_SERVERS <<< "$CONFIG"

    log_msg "Sanity Check - OS_TYPE: '$OS_TYPE', OS_VERSION: '$OS_VERSION'"

    case "$OS_TYPE" in
        "debian")
            IMAGE="debian/${OS_VERSION}/cloud/amd64"
            ;;
        "blueonyx")
            IMAGE="blueonyx/${OS_VERSION}/cloud"
            if ! fetch_blueonyx_image "$IMAGE"; then
                log_msg "Error: BlueOnyx image fetch failed. Cannot proceed."
                return 1
            fi
            ;;
        *)
            log_msg "Error: Unsupported OS type '$OS_TYPE'"
            return 1
            ;;
    esac

    log_msg "Using image $IMAGE for container creation."

    RAM_GIB=$(( RAM_MB / 1024 / 1024 ))GiB
    DISK_GIB=$(( DISK_MB / 1024 / 1024 ))GiB

    if incus list | grep -q "^| $CONTAINER_NAME"; then
        incus config device remove "$CONTAINER_NAME" root 2>/dev/null
        log_msg "RAM_GIB: $RAM_GIB"
        incus config set "$CONTAINER_NAME" limits.memory "$RAM_GIB"
        log_msg "DISK_GIB: $DISK_GIB"
        incus config device add "$CONTAINER_NAME" root disk path=/ pool=default size="$DISK_GIB"
        log_msg "Container $CONTAINER_NAME already exists. Updated configuration."
        return 0
    fi

    if ! incus image list | grep -q "$IMAGE"; then
        log_msg "Image $IMAGE not found locally. Attempting to copy..."
        incus image copy images:"$IMAGE" local: --alias "$IMAGE"
    fi

    log_msg "Initializing container $CONTAINER_NAME..."
    incus init "$IMAGE" "$CONTAINER_NAME" || return 1

    log_msg "Applying instance configuration for $CONTAINER_NAME..."

    incus config set "$CONTAINER_NAME" limits.memory "$RAM_GIB"
    incus config device override "$CONTAINER_NAME" root size="$DISK_GIB"

    incus config device override "$CONTAINER_NAME" eth0 nictype=bridged parent=vlan-br0
    incus config device set "$CONTAINER_NAME" eth0 vlan=100

    incus config set "$CONTAINER_NAME" user.instance_ipv4 "&${IP_ADDRESS}&"
    incus config set "$CONTAINER_NAME" user.instance_owner "default"

    NETWORK_CONFIG=$(cat <<EOF
version: 2
ethernets:
  eth0:
    dhcp4: no
    dhcp6: no
    ipv6: link-local
      - $IP_ADDRESS
    gateway4: $GATEWAY
    nameservers:
      addresses:
        - $(echo "$DNS_SERVERS" | sed 's/,/\n        - /g')
EOF
)
    incus config set "$CONTAINER_NAME" user.network-config "$NETWORK_CONFIG"

    USER_DATA=$(cat <<EOF
#cloud-config
hostname: $HOSTNAME
network:
  config: disabled
write_files:
  -
    path: /etc/systemd/network/eth0.network
    content: |-
      [Match]
      Name=eth0
      [Network]
      DNS=$(echo "$DNS_SERVERS" | sed 's/,/\n      DNS=/g')
      [Address]
      Address=$IP_ADDRESS
      [Route]
      Gateway=$GATEWAY
      Destination=0.0.0.0/0
      [IPv6AcceptRA]
      UseAutonomousPrefix=false
      [Network]
      LinkLocalAddressing=no
runcmd:
  - 'rm -f /etc/systemd/network/10-cloud-init-eth0.network'
  - 'rm -f /etc/NetworkManager/system-connections/10-cloud-init-eth0.network'
  - 'chmod 600 /etc/NetworkManager/system-connections/eth0.nmconnection'
  - 'chmod 644 /etc/systemd/network/eth0.network'
  - 'chown systemd-network:systemd-network /etc/systemd/network/eth0.network'
  - 'nmcli con reload'
  - 'systemctl condrestart systemd-networkd'
EOF
)
    incus config set "$CONTAINER_NAME" user.user-data "$USER_DATA"

    log_msg "Container $CONTAINER_NAME created and configured successfully."
}

# *** Sync container data with SSH check ***
sync_container_data() {
    CONTAINER_NAME=$1
    SOURCE_VPSID=$2
    SOURCE_SERVER=$3

    log_msg "Starting rsync for $CONTAINER_NAME..."

    check_ssh_connection "$SOURCE_SERVER"

    RSYNC_OPTIONS="-ah --info=progress2 --partial --delete --links --hard-links --perms --owner --group --devices --times"
    SSH_CMD="ssh -p $SSH_PORT -o Cipher=aes128-ctr -o Compression=no"
    
    SOURCE_PATH="root@$SOURCE_SERVER:/vz/root/$ENV_ID/"
    DEST_PATH="/home/incus/storage-pools/default/containers/$CONTAINER_NAME/rootfs/"

    rsync $RSYNC_OPTIONS -e "$SSH_CMD" "$SOURCE_PATH" "$DEST_PATH" || { 
        log_msg "Error: Rsync failed for $CONTAINER_NAME"
        return 1
    }

    rm -f "$DEST_PATH/etc/fstab"
}

# *** Enhanced from common.sh: Import process ***
backup_container() {
    log_msg "Starting import of OpenVZ 7 Container '$SOURCE_VPSID' from $SOURCE_SERVER as Incus Instance '$CONTAINER_NAME' ..."

    check_ssh_connection "$SOURCE_SERVER"

    INSTANCE_STATE=$(incus list --format=json | jq -r ".[] | select(.name == \"$CONTAINER_NAME\") | .status")
    if [[ "$INSTANCE_STATE" == "Running" ]]; then
        log_msg "Error: Container $CONTAINER_NAME is already running. Aborting import to prevent data corruption."
        return 1
    fi

    CONFIG=$(extract_openvz_config "$SOURCE_VPSID" 2>>"$LOG_FILE")
    if [ $? -ne 0 ]; then
        log_msg "Error: Failed to extract configuration for CT $SOURCE_VPSID - CT is not present, not running or not mounted."
        return 1
    fi

    create_container "$SOURCE_VPSID" "$CONTAINER_NAME"

    if ! incus list | grep -q "^| $CONTAINER_NAME"; then
        log_msg "Error: Container $CONTAINER_NAME does not exist. Skipping rsync."
        return 1
    fi

    sync_container_data "$CONTAINER_NAME" "$SOURCE_VPSID" "$SOURCE_SERVER"

    log_msg "Configuring network settings for $CONTAINER_NAME..."

    NETWORKS_PATH="/home/incus/storage-pools/default/containers/$CONTAINER_NAME/rootfs/etc/networks"
    INTERFACES_PATH="/home/incus/storage-pools/default/containers/$CONTAINER_NAME/rootfs/etc/network/interfaces"

    cat <<EOF > "$NETWORKS_PATH"
default         0.0.0.0
loopback        127.0.0.0
link-local      169.254.0.0
EOF
    log_msg "Updated /etc/networks for $CONTAINER_NAME."

    if [[ "$OS_TYPE" == "debian" ]]; then
        cat <<EOF > "$INTERFACES_PATH"
# This configuration file is auto-generated.
# WARNING: Do not edit this file, otherwise your changes will be lost.
# Please edit template /etc/network/interfaces.template instead.

auto lo
iface lo inet loopback

auto eth0
iface eth0 inet static
    address $IP_ADDRESS
    netmask $NETMASK
    gateway $GATEWAY
    dns-nameservers $(echo "$DNS_SERVERS" | tr ',' ' ')
EOF
        log_msg "Updated /etc/network/interfaces for $CONTAINER_NAME (Debian)."
    else
        log_msg "Skipping /etc/network/interfaces for $CONTAINER_NAME (Non-Debian)."
    fi

    if [[ "$OS_TYPE" == "blueonyx" ]]; then
        log_msg "Fixing BlueOnyx CODB IPType settings for $CONTAINER_NAME..."
        CODB_PATH="/home/incus/storage-pools/default/containers/$CONTAINER_NAME/rootfs/usr/sausalito/codb/objects"
        find "$CODB_PATH" -type f -name ".IPType" | while read -r iptype_file; do
            echo -n "IPv4" > "$iptype_file"
            log_msg "Updated: $iptype_file"
        done

        grep -rl "venet0:0" "$CODB_PATH"/*/.device | while read -r device_file; do
            echo -n "eth0" > "$device_file"
            log_msg "Updated: $device_file (venet0:0 → eth0)"
        done

        grep -rl "venet0" "$CODB_PATH"/*/.device | while read -r device_file; do
            echo -n "lo" > "$device_file"
            log_msg "Updated: $device_file (venet0 → lo)"
        done

        log_msg "Removing OpenVZ network scripts in $CONTAINER_NAME..."
        NETWORK_SCRIPTS_PATH="/home/incus/storage-pools/default/containers/$CONTAINER_NAME/rootfs/etc/sysconfig/network-scripts"
        find "$NETWORK_SCRIPTS_PATH" -type f -name "ifcfg-venet*" -exec rm -f {} \;

        log_msg "Creating /tmp/bx_cloud_cfg for BlueOnyx instance $CONTAINER_NAME..."
        INSTANCE_ROOTFS="/home/incus/storage-pools/default/containers/$CONTAINER_NAME/rootfs"
        BX_CLOUD_CFG="$INSTANCE_ROOTFS/tmp/bx_cloud_cfg"
        FQDN=$(cat "$INSTANCE_ROOTFS/etc/hostname" | tr -d '[:space:]')
        DNS_FORMATTED="&$(echo "$DNS_SERVERS" | tr ',' '&')&"
        cat <<EOF > "$BX_CLOUD_CFG"
FQDN|$FQDN
ipaddr|$IP_ADDRESS
netmask|$NETMASK
gateway|$GATEWAY
ipaddr_IPv6|
gateway_IPv6|
dns|$DNS_FORMATTED
dhcp4|no
dhcp6|no
EOF
        log_msg "Generated /tmp/bx_cloud_cfg for $CONTAINER_NAME."

        log_msg "Creating network configuration files for BlueOnyx instance $CONTAINER_NAME..."
        NETWORK_SCRIPTS_DIR="$INSTANCE_ROOTFS/etc/sysconfig/network-scripts"
        mkdir -p "$NETWORK_SCRIPTS_DIR"

        IFACE_CONFIG="$NETWORK_SCRIPTS_DIR/ifcfg-eth0"
        cat <<EOF > "$IFACE_CONFIG"
DEVICE=eth0
BOOTPROTO=none
ONBOOT=yes
DELAY=0
PROXY_METHOD=none
NM_CONTROLLED=yes
USERCTL=no
ARPCHECK=no
IPV6INIT=yes
IPV6ADDR=
IPV6_DEFAULTGW=
IPADDR=$IP_ADDRESS
NETMASK=$NETMASK
EOF
        IFS=',' read -ra DNS_ARRAY <<< "$DNS_SERVERS"
        for i in "${!DNS_ARRAY[@]}"; do
            echo "DNS$((i+1))=${DNS_ARRAY[$i]}" >> "$IFACE_CONFIG"
        done
        log_msg "Generated $IFACE_CONFIG."

        SYS_NETWORK="$INSTANCE_ROOTFS/etc/sysconfig/network"
        cat <<EOF > "$SYS_NETWORK"
FORWARD_IPV4=false
GATEWAY=$GATEWAY
HOSTNAME=$FQDN
NETWORKING=yes
NOZEROCONF=yes
EOF
        log_msg "Generated $SYS_NETWORK."
    fi

    log_msg "Setting SSH host key permissions for $CONTAINER_NAME..."
    SSH_KEYS=(
        "$INSTANCE_ROOTFS/etc/ssh/ssh_host_ecdsa_key"
        "$INSTANCE_ROOTFS/etc/ssh/ssh_host_rsa_key"
        "$INSTANCE_ROOTFS/etc/ssh/ssh_host_ed25519_key"
    )
    for KEY in "${SSH_KEYS[@]}"; do
        if [[ -f "$KEY" ]]; then
            chmod 600 "$KEY"
            chown root:root "$KEY"
            log_msg "Fixed permissions for $KEY"
        fi
    done

    log_msg "Removing quota files from $CONTAINER_NAME..."
    QUOTA_FILES=(
        "$INSTANCE_ROOTFS/aquota.group"
        "$INSTANCE_ROOTFS/aquota.user"
    )
    for FILE in "${QUOTA_FILES[@]}"; do
        if [[ -f "$FILE" ]]; then
            rm -f "$FILE"
            log_msg "Removed $FILE"
        fi
    done

    if [[ "$OS_TYPE" == "blueonyx" ]]; then
        log_msg "Ensuring NetworkManager is enabled in BlueOnyx instance $CONTAINER_NAME..."
        SYSTEMD_DIR="$INSTANCE_ROOTFS/etc/systemd/system"
        MULTI_USER_TARGET="$INSTANCE_ROOTFS/etc/systemd/system/multi-user.target.wants"
        NETWORKMANAGER_SERVICE="/usr/lib/systemd/system/NetworkManager.service"

        if [[ -f "$INSTANCE_ROOTFS$NETWORKMANAGER_SERVICE" ]]; then
            mkdir -p "$MULTI_USER_TARGET"
            ln -sf "$NETWORKMANAGER_SERVICE" "$MULTI_USER_TARGET/NetworkManager.service"
            log_msg "Enabled NetworkManager for auto-start in $CONTAINER_NAME."
        else
            log_msg "NetworkManager service file not found, skipping enablement."
        fi

        IFACE_CONFIG="$INSTANCE_ROOTFS/etc/sysconfig/network-scripts/ifcfg-eth0"
        if [[ -f "$IFACE_CONFIG" ]]; then
            sed -i 's/^NM_CONTROLLED=.*/NM_CONTROLLED=yes/' "$IFACE_CONFIG"
            log_msg "Updated NM_CONTROLLED=yes in $IFACE_CONFIG"
        fi
    fi

    log_msg "Import completed successfully for $CONTAINER_NAME."
}

# Execute the backup
backup_container

exit 0

# 
# Copyright (c) 2006-2025 Michael Stauber, SOLARSPEED.NET
# Copyright (c) 2006-2025 Team BlueOnyx, BLUEONYX.IT
# All Rights Reserved.
# 
# 1. Redistributions of source code must retain the above copyright 
#     notice, this list of conditions and the following disclaimer.
# 
# 2. Redistributions in binary form must reproduce the above copyright 
#     notice, this list of conditions and the following disclaimer in 
#     the documentation and/or other materials provided with the 
#     distribution.
# 
# 3. Neither the name of the copyright holder nor the names of its 
#     contributors may be used to endorse or promote products derived 
#     from this software without specific prior written permission.
# 
# THIS SOFTWARE IS PROVIDED BY THE COPYRIGHT HOLDERS AND CONTRIBUTORS 
# "AS IS" AND ANY EXPRESS OR IMPLIED WARRANTIES, INCLUDING, BUT NOT 
# LIMITED TO, THE IMPLIED WARRANTIES OF MERCHANTABILITY AND FITNESS 
# FOR A PARTICULAR PURPOSE ARE DISCLAIMED. IN NO EVENT SHALL THE 
# COPYRIGHT HOLDER OR CONTRIBUTORS BE LIABLE FOR ANY DIRECT, INDIRECT, 
# INCIDENTAL, SPECIAL, EXEMPLARY, OR CONSEQUENTIAL DAMAGES (INCLUDING, 
# BUT NOT LIMITED TO, PROCUREMENT OF SUBSTITUTE GOODS OR SERVICES; 
# LOSS OF USE, DATA, OR PROFITS; OR BUSINESS INTERRUPTION) HOWEVER 
# CAUSED AND ON ANY THEORY OF LIABILITY, WHETHER IN CONTRACT, STRICT 
# LIABILITY, OR TORT (INCLUDING NEGLIGENCE OR OTHERWISE) ARISING IN 
# ANY WAY OUT OF THE USE OF THIS SOFTWARE, EVEN IF ADVISED OF THE 
# POSSIBILITY OF SUCH DAMAGE.
# 
# You acknowledge that this software is not designed or intended for 
# use in the design, construction, operation or maintenance of any 
# nuclear facility.
# 
